"""Library with compatible bulb types.

Bulb Type detection:
ESP01_SHDW1C_31
ESP01 -- defines the module family (WiFi only bulb in this case)
SH -- Single Head light (most bulbs are single heads) / LED Strip
TW -- Tunable White - can only control CCT and dimming; no color
DW -- Dimmable White (most filament bulbs)
RGB -- Fullstack bulb
1C -- Specific to the hardware - defines PWM frequency + way of controlling CCT temperature
31 -- Related to the hardware revision
"""


from enum import Enum


class Features:
    """Defines the supported features."""

    color: bool
    color_tmp: bool
    effect: bool
    brightness: bool

    def __init__(
        self, color: bool, color_tmp: bool, effect: bool, brightness: bool
    ) -> None:
        """Init the features with type."""
        self.color = color
        self.color_tmp = color_tmp
        self.effect = effect
        self.brightness = brightness


class KelvinRange:
    """Defines the kelvin range."""

    max: int
    min: int

    def __init__(self, max: int, min: int) -> None:
        """Init for the kelvin range class."""
        self.max = max
        self.min = min


class BulbClass(Enum):
    """Bulb Types."""

    """Have Cool White and Warm White LEDs."""
    TW = "Tunable White"
    """Have only Dimmable white LEDs."""
    DW = "Dimmable White"
    """Have RGB LEDs."""
    RGB = "RGB Bulb"


class BulbType:
    """BulbType object to define functions and features of the bulb."""

    features: Features
    name: str
    filament_bulb: bool
    kelvin_range: KelvinRange
    bulb_type: BulbClass

    def __init__(
        self,
        features: Features,
        name: str,
        kelvin_range: KelvinRange,
        bulb_type: BulbClass,
    ) -> None:
        """Create a bulb object with different features."""
        self.features = features
        self.name = name
        self.kelvin_range = kelvin_range
        self.bulb_type = bulb_type



"""Discover bulbs in a network."""
import asyncio
import json
import logging
import socket

from pywizlight import wizlight

_LOGGER = logging.getLogger(__name__)


class DiscoveredBulb:
    """Representation of discovered bulb."""

    def __init__(self, ip_address, mac_address):
        """Initialize the discovered bulb."""
        self.ip_address = ip_address
        self.mac_address = mac_address

    @staticmethod
    def create_bulb_from_message(raw_addr, announce_message):
        """Create announce message."""
        ip_address = raw_addr[0]
        _LOGGER.debug("Found bulb with IP: {}".format(ip_address))
        bulb = DiscoveredBulb(
            ip_address=ip_address,
            mac_address=json.loads(announce_message)["result"]["mac"],
        )
        return bulb


class BulbRegistry(object):
    """Representation of the bulb registry."""

    def __init__(self):
        """Initialize the bulb registry."""
        self.bulbs_by_mac = {}

    def register(self, bulb):
        """Register a new bulb."""
        self.bulbs_by_mac[bulb.mac_address] = bulb

    def bulbs(self) -> list:
        """Get all present bulbs."""
        return list(self.bulbs_by_mac.values())


class BroadcastProtocol(object):
    """Protocol that sends an UDP broadcast message for bulb discovery."""

    def __init__(self, loop, registry, broadcast_address):
        """Init discovery function."""
        self.loop = loop
        self.registry = registry
        self.broadcast_address = broadcast_address

    def connection_made(self, transport):
        """Init connection to socket and register broadcasts."""
        self.transport = transport
        sock = transport.get_extra_info("socket")
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self.broadcast_registration()

    def broadcast_registration(self):
        """Send a registration method as UDP broadcast."""
        # Note: The IP and address we give the bulb here don't seem to matter for our
        # intents and purposes, so they're hardcoded to technically valid dummy data.
        # Fix for async problems if broaddcast_registration is called twice! See #13.
        # dirty dirty hack
        try:
            register_method = r'{"method":"registration","params":{"phoneMac":"AAAAAAAAAAAA","register":false,"phoneIp":"1.2.3.4","id":"1"}}'  # noqa: E501
            self.transport.sendto(
                register_method.encode(), (self.broadcast_address, 38899)
            )
            self.loop.call_later(1, self.broadcast_registration)
        except AttributeError:
            pass

    def datagram_received(self, data, addr):
        """Receive data from broadcast."""
        _LOGGER.debug("Received data {} from IP address {}".format(data, addr))
        decoded_message = data.decode()
        if """"success":true""" in decoded_message:
            bulb = DiscoveredBulb.create_bulb_from_message(addr, decoded_message)
            self.registry.register(bulb)

    def connection_lost(self, exc):
        """Return connection error."""
        _LOGGER.debug("Closing UDP discovery")


async def find_wizlights(wait_time=5, broadcast_address="255.255.255.255") -> list:
    """Start discovery and return list of IPs of the bulbs."""
    registry = BulbRegistry()
    loop = asyncio.get_event_loop()
    transport, protocol = await loop.create_datagram_endpoint(
        lambda: BroadcastProtocol(loop, registry, broadcast_address),
        local_addr=("0.0.0.0", 38899),
    )
    try:
        await asyncio.sleep(wait_time)
    finally:
        transport.close()
        bulbs = registry.bulbs()
        for bulb in bulbs:
            _LOGGER.info(
                "Discovered bulb {} with MAC {}".format(
                    bulb.ip_address, bulb.mac_address
                )
            )
        return bulbs


async def discover_lights(broadcast_space="255.255.255.255") -> list:
    """Find lights and return list with wizlight objects."""
    discovered_IPs = await find_wizlights(broadcast_address=broadcast_space)
    # empty list for adding bulbs
    bulbs = []
    # create light entities from register
    for entries in discovered_IPs:
        bulbs.append(wizlight(ip=entries.ip_address, mac=entries.mac_address))
    return bulbs



from pywizlight.bulb import PilotBuilder, PilotParser, wizlight  # noqa: 401

from pywizlight import discovery  # noqa: 401
from pywizlight.scenes import SCENES  # noqa: 401
from pywizlight.bulblibrary import BulbType  # noqa: 401


__all__ = ["wizlight"]



"""Definition of the supported bulb effects."""
SCENES = {
    1: "Ocean",
    2: "Romance",
    3: "Sunset",
    4: "Party",
    5: "Fireplace",
    6: "Cozy",
    7: "Forest",
    8: "Pastel Colors",
    9: "Wake up",
    10: "Bedtime",
    11: "Warm White",
    12: "Daylight",
    13: "Cool white",
    14: "Night light",
    15: "Focus",
    16: "Relax",
    17: "True colors",
    18: "TV time",
    19: "Plantgrowth",
    20: "Spring",
    21: "Summer",
    22: "Fall",
    23: "Deepdive",
    24: "Jungle",
    25: "Mojito",
    26: "Club",
    27: "Christmas",
    28: "Halloween",
    29: "Candlelight",
    30: "Golden white",
    31: "Pulse",
    32: "Steampunk",
    1000: "Rhythm",
}
